// Purpose:
//
//		This extension is designed to interact with Microsoft Visual Source
//		Safe version control system.
//
//		This code is based, in part, on a generic version control system
//		interface module written by John Kercheval.  Unlike that module,
//		this implementation interacts with an external windows DLL, so it
//		executes substantially faster, and provides more feedback than
//		shelling out to the command line version of VSS.
//
// Usage:
//
//		Compile and load this extension and save the state.
//
//		This extension is set up to allow the specification of a
//		default VSS project directory and it's associated project
//		name.
//
//		The VSS variables are stored in the session file to allow
//		project specific VSS settings.
//
// Details:
//
//		Before using any of these commands, you will be asked to set the
//		following VSS variables:
//				VSS_project_dir - local root directory of the VSS project
//				VSS_project_name - name of VSS project root folder
//
//		This module interacts with an external windows DLL ssiface.dll.
//		Place this DLL in the windows system directory, or add the appropriate
//		path in the code below.  The purpose of this DLL is to provide a
//		simple interface to the VSS COM interface.
//
//
//	Greg Brown
//  gbrown@iristech.com
//
//
// Modification History:
//
//		G. Brown 09/13/02 creation
//


#include "eel.h"

#define _A_RDONLY	0x01				// Read only file
#define _A_SUBDIR	0x10				// Subdirectory

user char VSS_project_dir[FNAMELEN];	// local root directory of VSS project
user char VSS_project_name[FNAMELEN];	// name of VSS project root folder

#define VSS_START_DELIM	"<<SES!@!VSS_START>>"
#define VSS_END_DELIM	"<<SES!@!VSS_END>>"

// may add DLL path here...
//#define SS_DLL "c:/dll_path/ssiface.dll"
#define SS_DLL "ssiface.dll"

#define SSCMD_CHECKOUT 0
#define SSCMD_CHECKIN 1
#define SSCMD_GET 2
#define SSCMD_ADD 3
#define SSCMD_CHECKOUT_STATUS 4
#define SSCMD_UNDO_CHECKOUT 5

new_vss_make_session_hook(b)
{
	save_var bufnum = b;

	point = size();

	buf_printf(b, "\n%s", VSS_START_DELIM);
	buf_printf(b, "\n%s", VSS_project_dir);
	buf_printf(b, "\n%s", VSS_project_name);
	buf_printf(b, "\n%s\n", VSS_END_DELIM);

	vss_make_session_hook(b);
}

REPLACE_FUNC("vss", "make_session_hook")

new_vss_restore_session_hook(b)
{
	int old_point = point;

	save_var bufnum = b;

	// obtain the project variables
	point = 0;
	if (!search(1, VSS_START_DELIM)) {
		point = old_point;
		} else {
		nl_forward();			 // move to first variable
		grab_line(b, VSS_project_dir);
		grab_line(b, VSS_project_name);
		nl_forward();			 // move beyond final delimiter
		}

	// call the previous hook
	vss_restore_session_hook(b);
}

REPLACE_FUNC("vss", "restore_session_hook")

int vss_is_writable(char *f)
{
	struct file_info fi;

	// check the file attribute, if it doesn't exist then it is writable
	if (!check_file(f, &fi))
		return 1;

	if (fi.attr & _A_RDONLY && !(fi.attr & _A_SUBDIR))
		return 0;
	return 1;
}

int vss_do_overwrite()
{
	char prompt[FNAMELEN];
	char response[80];
	
	if (vss_is_writable(filename))
		{
		sprintf(prompt, "%s is writeable, continue?", get_tail(filename, 0));
		get_strdef(response, prompt, "y");
		if (strnfcmp(response, "y", 1))
			return 0;
		}
	return 1;
}

int vss_get_file_name(char *f)
{
	int base_len;
	int name_len;

	name_len = strlen(VSS_project_name);
	base_len = strlen(VSS_project_dir);

	if (!name_len || !base_len)
		{
		say("Please set variables VSS_project_name and VSS_project_dir");
		return 0;
		}
	
	if (strncmp(VSS_project_dir, filename, base_len))
		{
		say("File %s does not reside in VSS project: %s.", filename, VSS_project_name);
		return 0;
		}

	sprintf(f, "$\\%s%s", VSS_project_name, &filename[base_len]);

	return 1;
}

void vss_update_file()
{
#if 0  // use the default behaviour of telling us that the file has changed...
	check_dates(0);
#else  // just reload file automatically
	int pos;
	refresh();
	pos = point;
	if (read_file(filename, reading_translation_type()))
		say("Error reading %s from disk.", filename);
	//else
	//	say("%s: Read from disk.", filename);
	point = pos;
	//to_begin_line();
#endif
}

int vss_command(char *f, int cmd, char *str, int str_len)
{
	return call_dll(SS_DLL, "sscmd", "c", "pipi", f, cmd, str, str_len);
}

command vss_checkin()
{
	char tmp_file[FNAMELEN];
	int rtrn;

	if (!vss_get_file_name(tmp_file))
		return;

	do_save_file(0, 1, 1);

	rtrn = vss_command(tmp_file, SSCMD_CHECKIN, NULL, 0);

	if (rtrn < 0)
		say("ERROR checking in %s", tmp_file);
	else
		{
		say("Checked in %s.", tmp_file);
		vss_update_file();
		}
}

command vss_checkout()
{
	char tmp_file[FNAMELEN];
	char user_name[FNAMELEN];
	int rtrn;

	if (!vss_get_file_name(tmp_file))
		return;

	rtrn = vss_command(tmp_file, SSCMD_CHECKOUT_STATUS, NULL, 0);
	if (rtrn == 2)
		{
		say("You already have %s checked out.", get_tail(filename, 0));
		return;
		}

	if (!vss_do_overwrite())
		return;

	rtrn = vss_command(tmp_file, SSCMD_CHECKOUT, user_name, sizeof(user_name));
	if (rtrn == 0)
		{
		say("Checked out %s.", tmp_file);
		vss_update_file();
		return;
		}
	else if (rtrn >= 1)
		{
		say("%s already checked out by %s.", tmp_file, user_name);
		return;
		}

	say("ERROR checking out %s", tmp_file);
}

command vss_get()
{
	char tmp_file[FNAMELEN];
	char prompt[FNAMELEN];
	char response[80];
	int rtrn;

	if (!vss_get_file_name(tmp_file))
		return;

	rtrn = vss_command(tmp_file, SSCMD_CHECKOUT_STATUS, NULL, 0);
	if (rtrn == 2)
		{
		sprintf(prompt, "You currently have %s checked out, overwrite?", get_tail(filename, 0));
		get_strdef(response, prompt, "y");
		if (strnfcmp(response, "y", 1))
			return;
		}
	else if (!vss_do_overwrite())
		return;

	rtrn = vss_command(tmp_file, SSCMD_GET, NULL, 0);

	if (rtrn < 0)
		say("ERROR getting %s", tmp_file);
	else
		{
		say("Got %s.", tmp_file);
		vss_update_file();
		}
}

command vss_add()
{
	char tmp_file[FNAMELEN];
	int rtrn;
	int i;
	int last_slash = -1;


	if (!vss_get_file_name(tmp_file))
		return;

	for (i = 0; tmp_file[i]; i++)
		{
		if (tmp_file[i] == '/' || tmp_file[i] == '\\')
			last_slash = i;
		}

	if (last_slash >= 0)
		tmp_file[last_slash] = 0;

	rtrn = vss_command(tmp_file, SSCMD_ADD, filename, 0);

	if (rtrn < 0)
		say("ERROR adding %s", tmp_file);
	else
		{
		say("Added %s to %s.", filename, tmp_file);
		vss_update_file();
		}
}

command vss_checkout_status()
{
	char tmp_file[FNAMELEN];
	char user_name[FNAMELEN];
	int rtrn;

	if (!vss_get_file_name(tmp_file))
		return;

	rtrn = vss_command(tmp_file, SSCMD_CHECKOUT_STATUS, user_name, sizeof(user_name));

	if (rtrn < 0)
		say("ERROR checking file status of FILE");
	else
		{
		if (rtrn == 0)
			say("%s not checked out.", tmp_file);
		else if (rtrn == 1 || rtrn == 2)
			say("%s checked out by %s.", tmp_file, user_name);
		else
			say("ERROR: unknown file status: %d.", rtrn);
		}
}

command vss_undo_checkout()
{
	char tmp_file[FNAMELEN];
	char prompt[FNAMELEN];
	char response[80];
	int rtrn;

	if (!vss_get_file_name(tmp_file))
		return;

	rtrn = vss_command(tmp_file, SSCMD_CHECKOUT_STATUS, NULL, 0);
	if (rtrn != 2)
		{
		say("You don't have %s checked out.", get_tail(filename, 0));
		return;
		}

	sprintf(prompt, "Undo checkout of %s?", get_tail(filename, 0));
	get_strdef(response, prompt, "y");
	if (strnfcmp(response, "y", 1))
		return;

	rtrn = vss_command(tmp_file, SSCMD_UNDO_CHECKOUT, NULL, 0);

	if (rtrn < 0)
		say("ERROR undoing check out of %s", tmp_file);
	else
		{
		say("Undid check out of %s.", tmp_file);
		vss_update_file();
		}
}
